﻿Module Module1

    Dim output As IO.StreamWriter

    Sub Main()
        If My.Application.CommandLineArgs.Count = 0 Then
            MsgBox("PackUCs - For NuPlayer" & vbCrLf & "by Yrex, 2019" & vbCrLf & vbCrLf & "Drag & drop the \Classes\ folder on this. All .uc files will be packed into (classes).uc. They can be deleted later.")
            Return
        End If

        Dim folder = My.Application.CommandLineArgs(0)
        Console.WriteLine("Packing files in " & folder & "...")

        output = New IO.StreamWriter(IO.Path.Combine(folder, "(classes).uc"))

        Dim files = IO.Directory.GetFiles(folder, "*.uc")
        For Each file In files
            If IO.Path.GetFileName(file).ToLower <> "(classes).uc" Then
                ProcessFile(file)
            End If
        Next

        output.Close()

        Console.WriteLine("Finished. Press any key...")
        Console.ReadKey(True)
    End Sub

    Enum EPosition
        BeforeClassName
        AfterClassName
        AfterParentName
        Props
    End Enum

    Sub ProcessFile(file As String)
        Using r As New IO.StreamReader(file)
            Dim pos As EPosition
            Dim className = "", parentName = ""
            Dim hasWrittenClassName As Boolean

            Do
                Dim ln = r.ReadLine
                If ln <> "" Then ln = ln.Trim()

                If pos < EPosition.Props Then
                    If ln.StartsWith("#exec", StringComparison.OrdinalIgnoreCase) Then
                        Dim ln2 = ln.Substring(5).Trim()

                        If ln2.StartsWith("mesh", StringComparison.OrdinalIgnoreCase) Then
                            If Not hasWrittenClassName Then
                                output.WriteLine("class " & className & " extends " & parentName & ";")
                                hasWrittenClassName = True
                            End If
                            output.WriteLine(ln)
                        End If
                    End If
                End If

                If pos = EPosition.BeforeClassName Then
                    If ln.StartsWith("class", StringComparison.OrdinalIgnoreCase) Then
                        className = GetStringAfter(ln, "class ", " ")
                        pos = EPosition.AfterClassName

                        parentName = GetStringAfter(ln, "extends ", "; ")
                        If parentName <> "" Then
                            pos = EPosition.AfterParentName
                        Else
                            parentName = GetStringAfter(ln, "expands ", "; ")
                            If parentName <> "" Then pos = EPosition.AfterParentName
                        End If
                    End If
                End If

                If pos = EPosition.AfterClassName Then
                    If ln.StartsWith("extends", StringComparison.OrdinalIgnoreCase) OrElse
                       ln.StartsWith("expands", StringComparison.OrdinalIgnoreCase) Then
                        parentName = GetStringAfter(ln, "extends ", "; ")
                        If parentName = "" Then parentName = GetStringAfter(ln, "expands ", "; ")
                        If parentName <> "" Then pos = EPosition.AfterParentName
                    End If
                End If

                If pos < EPosition.Props Then
                    If String.Equals(ln, "defaultproperties", StringComparison.OrdinalIgnoreCase) Then
                        If Not hasWrittenClassName Then
                            output.WriteLine("class " & className & " extends " & parentName & ";")
                            hasWrittenClassName = True
                        End If
                        pos = EPosition.Props
                        output.WriteLine(ln)
                        output.WriteLine("{")
                    End If
                End If

                If pos = EPosition.Props Then
                    If ln.Contains("=") Then
                        output.WriteLine(ln)
                    End If
                End If

            Loop Until r.EndOfStream
            output.WriteLine("}")
        End Using
    End Sub

    Function GetStringPart(str As String, begin As Integer, endd As Integer) As String
        Return str.Substring(begin, endd - begin + 1)
    End Function

    Function GetStringAfter(str As String, part As String, endChar As Char) As String
        Dim strstart As Integer
        Dim strend As Integer
        Dim partIndex = str.IndexOf(part, StringComparison.OrdinalIgnoreCase)
        If partIndex = -1 Then Return ""
        strstart = partIndex + part.Length
        Dim endCharPosition = IndexOfAny(str, endChar, strstart)

        If endCharPosition <> -1 Then
            strend = endCharPosition - 1
        Else
            'if end char not found then end where string ends
            strend = str.Length - 1
        End If

        Return GetStringPart(str, strstart, strend)
    End Function

    Function IndexOfAny(str As String, chars As String, startIndex As Integer) As Integer
        For i = 0 To chars.Length - 1
            Dim index = str.IndexOf(chars(i).ToString(), startIndex, StringComparison.OrdinalIgnoreCase)
            If index <> -1 Then Return index
        Next
        Return -1
    End Function

End Module
